from enum import Enum
from typing import List, Optional, Union
from pydantic import BaseModel

class Providers(str, Enum):
    OPENAI = 'openai'
    ANTHROPIC = 'anthropic'
    OPENAI_ASSISTANTS = 'openai_assistants'
    VERCEL = 'vercel'
    GENTORO = 'gentoro'

class AuthenticationScope(str, Enum):
    METADATA = 'metadata'
    API_KEY = 'api_key'

class BaseObject(BaseModel):
    pass

class ScopeForMetadata(BaseObject):
    key_name: str

class Authentication(BaseObject):
    scope: AuthenticationScope
    metadata: Optional[ScopeForMetadata]

class SdkConfig(BaseObject):
    base_url: Optional[str]
    auth_mod_base_url: Optional[str]
    timeout_ms: Optional[int]
    api_key: Optional[str]
    provider: Optional[Providers]
    authentication: Optional[Authentication]

class Request(BaseObject):
    uri: str
    content: BaseObject

class Response(BaseObject):
    content: BaseObject

class Message(BaseObject):
    role: str
    content: str

class Context(BaseObject):
    bridge_uid: str
    messages: Optional[List[Message]]

class KeyValuePair(BaseObject):
    key: str
    value: str

class GetToolsRequest(BaseObject):
    context: Context
    metadata: List[KeyValuePair]

class FunctionParameter(BaseObject):
    name: str
    type: str
    description: str

class FunctionParameterCollection(BaseObject):
    properties: List[FunctionParameter]
    required: List[str]

class Function(BaseObject):
    name: str
    description: str
    parameters: FunctionParameterCollection

class ToolDef(BaseObject):
    type: str
    definition: Function

class GetToolsResponse(BaseObject):
    tools: List[ToolDef]

class TextContent(BaseObject):
    text: str

class NumberContent(BaseObject):
    number: float

class BoolContent(BaseObject):
    flag: bool

class DataType(str, Enum):
    STRING = 'string'
    NUMBER = 'number'
    OBJECT = 'object'
    BOOLEAN = 'boolean'
    ARRAY = 'array'

class DataValue(BaseObject):
    field_name: str
    data_type: DataType
    value: Union[TextContent, NumberContent, BoolContent]

class ArrayContent(BaseObject):
    entries: List[DataValue]

class ObjectContent(BaseObject):
    object: DataValue

class AuthenticationData(BaseObject):
    type: str
    connection_uid: str
    tool_call_id: str
    request_uid: str
    request_secret: str
    values: List[DataValue]

class FunctionCall(BaseObject):
    name: str
    arguments: str

class ToolCall(BaseObject):
    id: str
    type: str
    details: FunctionCall

class RunToolsRequest(BaseObject):
    context: Context
    authentication: Authentication
    metadata: List[KeyValuePair]
    tool_calls: List[ToolCall]

class ExecResultType(str, Enum):
    EXEC_OUTPUT = 'exec_output'
    ERROR = 'error'
    AUTH_REQUEST = 'auth_request'

class ExecOutput(BaseObject):
    content_type: str
    content: str

class ExecError(BaseObject):
    code: str
    message: str
    details: str

class AuthSchemaField(BaseObject):
    name: str
    description: str
    data_type: DataType
    fields: Optional[dict]

class AuthSchema(BaseObject):
    fields: List[AuthSchemaField]

class AuthenticationType(str, Enum):
    OAUTH = 'oauth'
    BASIC = 'basic'
    KEYPAIR = 'keypair'
    KEY = 'key'
    API_KEY = 'apikey'
    JWT = 'jwt'
    UNKNOWN = 'unknown'

class AuthRequest(BaseObject):
    connection_uid: str
    request_uid: str
    request_secret: str
    type: AuthenticationType
    settings: Optional[List[KeyValuePair]]
    schema: Optional[AuthSchema]

class AuthRequests(BaseObject):
    requests: List[AuthRequest]

class ExecResult(BaseObject):
    type: ExecResultType
    tool_call_id: str
    tool_uid: Optional[str]
    data: Union[ExecOutput, ExecError, AuthRequests]

class RunToolsResponse(BaseObject):
    results: List[ExecResult]

class SdkError(BaseObject):
    code: str
    message: str
    details: str

class SdkEventType(str, Enum):
    AUTHENTICATION_REQUEST = 'authentication_request'

class AuthenticationStatus(str, Enum):
    REQUESTED = 'requested'
    AUTHENTICATED = 'authenticated'
    EXPIRED = 'expired'
    ERROR = 'error'

class SdkAuthenticationEventInfo(BaseObject):
    tool_call_id: str
    tool_uid: str
    auth_request: AuthRequest

class SdkEvent(BaseObject):
    event_type: SdkEventType
    sdk: dict
    event_info: Union[SdkAuthenticationEventInfo]

class GetAuthStatusRequest(BaseObject):
    connection_uid: str
    request_uid: str
    request_secret: str

class AuthStatusSuccess(BaseObject):
    type: AuthenticationType
    authenticated_at: float

class AuthStatusError(BaseObject):
    code: str
    message: str
    stack_trace: str

class GetAuthStatusResponse(BaseObject):
    result: AuthenticationStatus
    info: Optional[Union[AuthStatusError, AuthStatusSuccess]]




from typing import List, Optional, Dict
from enum import Enum
import requests
import json


class Providers(str, Enum):
    OPENAI = 'openai'
    ANTHROPIC = 'anthropic'
    OPENAI_ASSISTANTS = 'openai_assistants'
    VERCEL = 'vercel'
    GENTORO = 'gentoro'


class AuthenticationScope(str, Enum):
    METADATA = 'metadata'
    API_KEY = 'api_key'


class Authentication:
    def __init__(self, scope: AuthenticationScope, metadata: Optional[Dict] = None):
        self.scope = scope
        self.metadata = metadata


class SdkConfig:
    def __init__(self, base_url: str, auth_mod_base_url: str, api_key: str, provider: Providers,
                 authentication: Authentication):
        if not api_key:
            raise ValueError("The api_key client option must be set")
        if not auth_mod_base_url:
            raise ValueError("Authentication module base URL is required")

        self.base_url = base_url
        self.auth_mod_base_url = auth_mod_base_url
        self.api_key = api_key
        self.provider = provider
        self.authentication = authentication


class Transport:
    def __init__(self, config: SdkConfig):
        self.config = config

    def send_request(self, uri: str, content: Dict, method: str = "POST", headers: Dict = None):
        url = f"{self.config.base_url}{uri}"

        if headers is None:
            headers = {
                "X-API-Key": self.config.api_key,
                "Accept": "application/json"
            }

        try:
            if method.upper() == "GET":
                response = requests.get(url, headers=headers, timeout=10)
            else:
                response = requests.post(url, json=content, headers=headers, timeout=10)

            response.raise_for_status()
            return response.json()

        except requests.exceptions.RequestException as e:
            print(f"Request failed: {e}")
            return None


class Gentoro:
    def __init__(self, config: SdkConfig, metadata: List[Dict] = None):
        self.transport = Transport(config)
        self.auth_mod_uri = config.auth_mod_base_url
        self.authentication = config.authentication
        self.metadata = metadata or []
        self.auth_request_checker_id = None
        self.config = config

    def metadata(self, key: str, value: str):
        self.metadata.append({"key": key, "value": value})
        return self

    def get_tools(self, bridge_uid: str, messages: Optional[List[Dict]] = None):
        try:
            request_uri = f"/api/bornio/v1/inference/{bridge_uid}/retrievetools"

            headers = {
                "X-API-Key": self.config.api_key,
                "Accept": "application/json",
                "User-Agent": "Python-SDK"
            }

            request_content = {
                "context": {"bridgeUid": bridge_uid, "messages": messages or []},
                "metadata": self.metadata
            }

            result = self.transport.send_request(request_uri, request_content, headers=headers, method="POST")

            if result and "tools" in result:
                return self._as_provider_tools(result["tools"])
            return None
        except Exception as e:
            print(f"Error fetching tools: {e}")
            return None

    def _as_provider_tools(self, tools: List[Dict]) -> List[Dict]:
        if self.config.provider == Providers.OPENAI:
            return [
                {
                    "type": "function",
                    "function": {
                        "name": tool["definition"]["name"],
                        "description": tool["definition"]["description"],
                        "parameters": {
                            "type": "object",
                            "properties": {
                                param["name"]: {"type": param["type"], "description": param["description"]}
                                for param in tool["definition"]["parameters"].get("properties", [])
                            },
                            "required": tool["definition"]["parameters"].get("required", []),
                        },
                    },
                }
                for tool in tools
            ]
        return tools

    def as_internal_tool_calls(self, messages: Dict) -> Optional[List[Dict]]:
        if self.config.provider == Providers.OPENAI:
            if "choices" in messages and messages["choices"][0].get("finish_reason") == "tool_calls":
                tool_calls = messages["choices"][0].get("message", {}).get("tool_calls", [])
                return [
                    {
                        "id": call["id"],
                        "type": call["type"],
                        "details": {
                            "name": call["function"]["name"],
                            "arguments": call["function"]["arguments"]
                        }
                    }
                    for call in tool_calls
                ]
        return None


    def run_tools(self, bridge_uid: str, messages: Optional[List[Dict]], tool_calls: List[Dict]):
        try:
            request_uri = f"/api/bornio/v1/inference/{bridge_uid}/runtools"

            headers = {
                "X-API-Key": self.config.api_key,
                "Accept": "application/json",
                "User-Agent": "Python-SDK"
            }

            extracted_tool_calls = self.as_internal_tool_calls(messages)  # <-- Calling as_internal_tool_calls here
            if extracted_tool_calls:
                tool_calls.extend(extracted_tool_calls)

            for tool_call in tool_calls:
                if "details" in tool_call and "arguments" in tool_call["details"]:
                    tool_call["details"]["arguments"] = json.dumps(tool_call["details"].get("arguments", {}))

            request_content = {
                "context": {"bridgeUid": bridge_uid, "messages": messages or []},
                "metadata": self.metadata,
                "authentication": {
                    "scope": self.authentication.scope.value,
                    "metadata": self.authentication.metadata if self.authentication.metadata else None
                },
                "toolCalls": tool_calls
            }

            result = self.transport.send_request(request_uri, request_content, headers=headers, method="POST")

            if result and "results" in result:
                return result["results"]
            return None
        except Exception as e:
            print(f"Error running tools: {e}")
            return None


    def add_event_listener(self, event_type: str, handler):
        try:
            print(f"Adding event listener for {event_type}")
        except Exception as e:
            print(f"Error adding event listener: {e}")


import setuptools


def read_requirements():
    with open("requirements.txt", encoding="utf-8") as req_file:
        return req_file.read().splitlines()


setuptools.setup(
    name="Gentoro",
    version="0.1.2",
    author="Gentoro R&D",
    author_email="communitysupport@gentoro.com",
    description="Gentoro Python SDK for AI tool execution and authentication",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/gentoro-GT/python-sdk",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=read_requirements(),
    include_package_data=True,
    license="Apache-2.0",
)
